namespace BitBadger.Documents.Postgres

/// <summary>Commands to execute custom SQL queries</summary>
[<RequireQualifiedAccess>]
module Custom =

    /// <summary>Execute a query that returns a list of results</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function between the document and the domain item</param>
    /// <returns>A list of results for the given query</returns>
    [<CompiledName "FSharpList">]
    let list<'TDoc> query parameters (mapFunc: RowReader -> 'TDoc) =
        WithProps.Custom.list<'TDoc> query parameters mapFunc (fromDataSource ())

    /// <summary>Execute a query that returns a list of results</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function between the document and the domain item</param>
    /// <returns>A list of results for the given query</returns>
    let List<'TDoc>(query, parameters, mapFunc: System.Func<RowReader, 'TDoc>) =
        WithProps.Custom.List<'TDoc>(query, parameters, mapFunc, fromDataSource ())

    /// <summary>Execute a query that returns a JSON array of results</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    /// <returns>A JSON array of results for the given query</returns>
    [<CompiledName "FSharpJsonArray">]
    let jsonArray query parameters mapFunc =
        WithProps.Custom.jsonArray query parameters mapFunc (fromDataSource ())

    /// <summary>Execute a query that returns a JSON array of results</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    /// <returns>A JSON array of results for the given query</returns>
    let JsonArray(query, parameters, mapFunc) =
        WithProps.Custom.JsonArray(query, parameters, mapFunc, fromDataSource ())

    /// <summary>Execute a query, writing its results to the given <c>StreamWriter</c></summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    [<CompiledName "FSharpWriteJsonArray">]
    let writeJsonArray query parameters writer mapFunc =
        WithProps.Custom.writeJsonArray query parameters writer mapFunc (fromDataSource ())

    /// <summary>Execute a query, writing its results to the given <c>StreamWriter</c></summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    let WriteJsonArray(query, parameters, writer, mapFunc) =
        WithProps.Custom.WriteJsonArray(query, parameters, writer, mapFunc, fromDataSource ())

    /// <summary>Execute a query that returns one or no results</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function between the document and the domain item</param>
    /// <returns><c>Some</c> with the first matching result, or <c>None</c> if not found</returns>
    [<CompiledName "FSharpSingle">]
    let single<'TDoc> query parameters (mapFunc: RowReader ->  'TDoc) =
        WithProps.Custom.single<'TDoc> query parameters mapFunc (fromDataSource ())

    /// <summary>Execute a query that returns one or no results</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function between the document and the domain item</param>
    /// <returns>The first matching result, or <c>null</c> if not found</returns>
    let Single<'TDoc when 'TDoc: null and 'TDoc: not struct>(
            query, parameters, mapFunc: System.Func<RowReader, 'TDoc>) =
        WithProps.Custom.Single<'TDoc>(query, parameters, mapFunc, fromDataSource ())

    /// <summary>Execute a query that returns one or no JSON documents</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    /// <returns>The JSON document with the first matching result, or an empty document if not found</returns>
    [<CompiledName "FSharpJsonSingle">]
    let jsonSingle query parameters mapFunc =
        WithProps.Custom.jsonSingle query parameters mapFunc (fromDataSource ())

    /// <summary>Execute a query that returns one or no JSON documents</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    /// <returns>The JSON document with the first matching result, or an empty document if not found</returns>
    let JsonSingle(query, parameters, mapFunc) =
        WithProps.Custom.JsonSingle(query, parameters, mapFunc, fromDataSource ())

    /// <summary>Execute a query that returns no results</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    [<CompiledName "NonQuery">]
    let nonQuery query parameters =
        WithProps.Custom.nonQuery query parameters (fromDataSource ())

    /// <summary>Execute a query that returns a scalar value</summary>
    /// <param name="query">The query to retrieve the value</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to obtain the value</param>
    /// <returns>The scalar value for the query</returns>
    [<CompiledName "FSharpScalar">]
    let scalar<'T when 'T: struct> query parameters (mapFunc: RowReader -> 'T) =
        WithProps.Custom.scalar query parameters mapFunc (fromDataSource ())

    /// <summary>Execute a query that returns a scalar value</summary>
    /// <param name="query">The query to retrieve the value</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to obtain the value</param>
    /// <returns>The scalar value for the query</returns>
    let Scalar<'T when 'T: struct>(query, parameters, mapFunc: System.Func<RowReader, 'T>) =
        WithProps.Custom.Scalar<'T>(query, parameters, mapFunc, fromDataSource ())


/// <summary>Table and index definition commands</summary>
[<RequireQualifiedAccess>]
module Definition =

    /// <summary>Create a document table</summary>
    /// <param name="name">The table whose existence should be ensured (may include schema)</param>
    [<CompiledName "EnsureTable">]
    let ensureTable name =
        WithProps.Definition.ensureTable name (fromDataSource ())

    /// <summary>Create an index on documents in the specified table</summary>
    /// <param name="name">The table to be indexed (may include schema)</param>
    /// <param name="idxType">The type of document index to create</param>
    [<CompiledName "EnsureDocumentIndex">]
    let ensureDocumentIndex name idxType =
        WithProps.Definition.ensureDocumentIndex name idxType (fromDataSource ())

    /// <summary>Create an index on field(s) within documents in the specified table</summary>
    /// <param name="tableName">The table to be indexed (may include schema)</param>
    /// <param name="indexName">The name of the index to create</param>
    /// <param name="fields">One or more fields to be indexed</param>
    [<CompiledName "EnsureFieldIndex">]
    let ensureFieldIndex tableName indexName fields =
        WithProps.Definition.ensureFieldIndex tableName indexName fields (fromDataSource ())


/// <summary>Document writing functions</summary>
[<AutoOpen>]
module Document =

    /// <summary>Insert a new document</summary>
    /// <param name="tableName">The table into which the document should be inserted (may include schema)</param>
    /// <param name="document">The document to be inserted</param>
    [<CompiledName "Insert">]
    let insert<'TDoc> tableName (document: 'TDoc) =
        WithProps.Document.insert<'TDoc> tableName document (fromDataSource ())

    /// <summary>Save a document, inserting it if it does not exist and updating it if it does (AKA "upsert")</summary>
    /// <param name="tableName">The table into which the document should be saved (may include schema)</param>
    /// <param name="document">The document to be saved</param>
    [<CompiledName "Save">]
    let save<'TDoc> tableName (document: 'TDoc) =
        WithProps.Document.save<'TDoc> tableName document (fromDataSource ())


/// <summary>Queries to count documents</summary>
[<RequireQualifiedAccess>]
module Count =

    /// <summary>Count all documents in a table</summary>
    /// <param name="tableName">The table in which documents should be counted (may include schema)</param>
    /// <returns>The count of the documents in the table</returns>
    [<CompiledName "All">]
    let all tableName =
        WithProps.Count.all tableName (fromDataSource ())

    /// <summary>Count matching documents using JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table in which documents should be counted (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>The count of matching documents in the table</returns>
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields =
        WithProps.Count.byFields tableName howMatched fields (fromDataSource ())

    /// <summary>Count matching documents using a JSON containment query (<c>@&gt;</c>)</summary>
    /// <param name="tableName">The table in which documents should be counted (may include schema)</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <returns>The count of the documents in the table</returns>
    [<CompiledName "ByContains">]
    let byContains tableName criteria =
        WithProps.Count.byContains tableName criteria (fromDataSource ())

    /// <summary>Count matching documents using a JSON Path match query (<c>@?</c>)</summary>
    /// <param name="tableName">The table in which documents should be counted (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to be matched</param>
    /// <returns>The count of the documents in the table</returns>
    [<CompiledName "ByJsonPath">]
    let byJsonPath tableName jsonPath =
        WithProps.Count.byJsonPath tableName jsonPath (fromDataSource ())


/// <summary>Queries to determine if documents exist</summary>
[<RequireQualifiedAccess>]
module Exists =

    /// <summary>Determine if a document exists for the given ID</summary>
    /// <param name="tableName">The table in which existence should be checked (may include schema)</param>
    /// <param name="docId">The ID of the document whose existence should be checked</param>
    /// <returns>True if a document exists, false if not</returns>
    [<CompiledName "ById">]
    let byId tableName docId =
        WithProps.Exists.byId tableName docId (fromDataSource ())

    /// <summary>Determine if a document exists using JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table in which existence should be checked (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>True if any matching documents exist, false if not</returns>
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields =
        WithProps.Exists.byFields tableName howMatched fields (fromDataSource ())

    /// <summary>Determine if a document exists using a JSON containment query (<c>@&gt;</c>)</summary>
    /// <param name="tableName">The table in which existence should be checked (may include schema)</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <returns>True if any matching documents exist, false if not</returns>
    [<CompiledName "ByContains">]
    let byContains tableName criteria =
        WithProps.Exists.byContains tableName criteria (fromDataSource ())

    /// <summary>Determine if a document exists using a JSON Path match query (<c>@?</c>)</summary>
    /// <param name="tableName">The table in which existence should be checked (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to be matched</param>
    /// <returns>True if any matching documents exist, false if not</returns>
    [<CompiledName "ByJsonPath">]
    let byJsonPath tableName jsonPath =
        WithProps.Exists.byJsonPath tableName jsonPath (fromDataSource ())


/// <summary>Commands to retrieve documents as domain objects</summary>
[<RequireQualifiedAccess>]
module Find =

    /// <summary>Retrieve all documents in the given table</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <returns>All documents from the given table</returns>
    [<CompiledName "FSharpAll">]
    let all<'TDoc> tableName =
        WithProps.Find.all<'TDoc> tableName (fromDataSource ())

    /// <summary>Retrieve all documents in the given table</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <returns>All documents from the given table</returns>
    let All<'TDoc> tableName =
        WithProps.Find.All<'TDoc>(tableName, fromDataSource ())

    /// <summary>Retrieve all documents in the given table ordered by the given fields in the document</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents from the given table, ordered by the given fields</returns>
    [<CompiledName "FSharpAllOrdered">]
    let allOrdered<'TDoc> tableName orderFields =
        WithProps.Find.allOrdered<'TDoc> tableName orderFields (fromDataSource ())

    /// <summary>Retrieve all documents in the given table ordered by the given fields in the document</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents from the given table, ordered by the given fields</returns>
    let AllOrdered<'TDoc> tableName orderFields =
        WithProps.Find.AllOrdered<'TDoc>(tableName, orderFields, fromDataSource ())

    /// <summary>Retrieve a document by its ID</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="docId">The ID of the document to retrieve</param>
    /// <returns><c>Some</c> with the document if found, <c>None</c> otherwise</returns>
    [<CompiledName "FSharpById">]
    let byId<'TKey, 'TDoc> tableName docId =
        WithProps.Find.byId<'TKey, 'TDoc> tableName docId (fromDataSource ())

    /// <summary>Retrieve a document by its ID</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="docId">The ID of the document to retrieve</param>
    /// <returns>The document if found, <c>null</c> otherwise</returns>
    let ById<'TKey, 'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, docId: 'TKey) =
        WithProps.Find.ById<'TKey, 'TDoc>(tableName, docId, fromDataSource ())

    /// <summary>Retrieve documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>All documents matching the given fields</returns>
    [<CompiledName "FSharpByFields">]
    let byFields<'TDoc> tableName howMatched fields =
        WithProps.Find.byFields<'TDoc> tableName howMatched fields (fromDataSource ())

    /// <summary>Retrieve documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>All documents matching the given fields</returns>
    let ByFields<'TDoc>(tableName, howMatched, fields) =
        WithProps.Find.ByFields<'TDoc>(tableName, howMatched, fields, fromDataSource ())

    /// <summary>
    /// Retrieve documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given fields in the
    /// document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents matching the given fields, ordered by the other given fields</returns>
    [<CompiledName "FSharpByFieldsOrdered">]
    let byFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields =
        WithProps.Find.byFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields (fromDataSource ())

    /// <summary>
    /// Retrieve documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given fields in the
    /// document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents matching the given fields, ordered by the other given fields</returns>
    let ByFieldsOrdered<'TDoc>(tableName, howMatched, queryFields, orderFields) =
        WithProps.Find.ByFieldsOrdered<'TDoc>(tableName, howMatched, queryFields, orderFields, fromDataSource ())

    /// <summary>Retrieve documents matching a JSON containment query (<c>@&gt;</c>)</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <returns>All documents matching the given containment query</returns>
    [<CompiledName "FSharpByContains">]
    let byContains<'TDoc> tableName (criteria: obj) =
        WithProps.Find.byContains<'TDoc> tableName criteria (fromDataSource ())

    /// <summary>Retrieve documents matching a JSON containment query (<c>@&gt;</c>)</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <returns>All documents matching the given containment query</returns>
    let ByContains<'TDoc>(tableName, criteria: obj) =
        WithProps.Find.ByContains<'TDoc>(tableName, criteria, fromDataSource ())

    /// <summary>
    /// Retrieve documents matching a JSON containment query (<c>@&gt;</c>) ordered by the given fields in the document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents matching the given containment query, ordered by the given fields</returns>
    [<CompiledName "FSharpByContainsOrdered">]
    let byContainsOrdered<'TDoc> tableName (criteria: obj) orderFields =
        WithProps.Find.byContainsOrdered<'TDoc> tableName criteria orderFields (fromDataSource ())

    /// <summary>
    /// Retrieve documents matching a JSON containment query (<c>@&gt;</c>) ordered by the given fields in the document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents matching the given containment query, ordered by the given fields</returns>
    let ByContainsOrdered<'TDoc>(tableName, criteria: obj, orderFields) =
        WithProps.Find.ByContainsOrdered<'TDoc>(tableName, criteria, orderFields, fromDataSource ())

    /// <summary>Retrieve documents matching a JSON Path match query (<c>@?</c>)</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <returns>All documents matching the given JSON Path expression</returns>
    [<CompiledName "FSharpByJsonPath">]
    let byJsonPath<'TDoc> tableName jsonPath =
        WithProps.Find.byJsonPath<'TDoc> tableName jsonPath (fromDataSource ())

    /// <summary>Retrieve documents matching a JSON Path match query (<c>@?</c>)</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <returns>All documents matching the given JSON Path expression</returns>
    let ByJsonPath<'TDoc>(tableName, jsonPath) =
        WithProps.Find.ByJsonPath<'TDoc>(tableName, jsonPath, fromDataSource ())

    /// <summary>
    /// Retrieve documents matching a JSON Path match query (<c>@?</c>) ordered by the given fields in the document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents matching the given JSON Path expression, ordered by the given fields</returns>
    [<CompiledName "FSharpByJsonPathOrdered">]
    let byJsonPathOrdered<'TDoc> tableName jsonPath orderFields =
        WithProps.Find.byJsonPathOrdered<'TDoc> tableName jsonPath orderFields (fromDataSource ())

    /// <summary>
    /// Retrieve documents matching a JSON Path match query (<c>@?</c>) ordered by the given fields in the document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents matching the given JSON Path expression, ordered by the given fields</returns>
    let ByJsonPathOrdered<'TDoc>(tableName, jsonPath, orderFields) =
        WithProps.Find.ByJsonPathOrdered<'TDoc>(tableName, jsonPath, orderFields, fromDataSource ())

    /// <summary>Retrieve the first document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns><c>Some</c> with the first document, or <c>None</c> if not found</returns>
    [<CompiledName "FSharpFirstByFields">]
    let firstByFields<'TDoc> tableName howMatched fields =
        WithProps.Find.firstByFields<'TDoc> tableName howMatched fields (fromDataSource ())

    /// <summary>Retrieve the first document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>The first document, or <c>null</c> if not found</returns>
    let FirstByFields<'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, howMatched, fields) =
        WithProps.Find.FirstByFields<'TDoc>(tableName, howMatched, fields, fromDataSource ())

    /// <summary>
    /// Retrieve the first document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given
    /// fields in the document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns><c>Some</c> with the first document ordered by the given fields, or <c>None</c> if not found</returns>
    [<CompiledName "FSharpFirstByFieldsOrdered">]
    let firstByFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields =
        WithProps.Find.firstByFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields (fromDataSource ())

    /// <summary>
    /// Retrieve the first document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given
    /// fields in the document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>The first document ordered by the given fields, or <c>null</c> if not found</returns>
    let FirstByFieldsOrdered<'TDoc when 'TDoc: null and 'TDoc: not struct>(
            tableName, howMatched, queryFields, orderFields) =
        WithProps.Find.FirstByFieldsOrdered<'TDoc>(tableName, howMatched, queryFields, orderFields, fromDataSource ())

    /// <summary>Retrieve the first document matching a JSON containment query (<c>@&gt;</c>)</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <returns><c>Some</c> with the first document, or <c>None</c> if not found</returns>
    [<CompiledName "FSharpFirstByContains">]
    let firstByContains<'TDoc> tableName (criteria: obj) =
        WithProps.Find.firstByContains<'TDoc> tableName criteria (fromDataSource ())

    /// <summary>Retrieve the first document matching a JSON containment query (<c>@&gt;</c>)</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <returns>The first document, or <c>null</c> if not found</returns>
    let FirstByContains<'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, criteria: obj) =
        WithProps.Find.FirstByContains<'TDoc>(tableName, criteria, fromDataSource ())

    /// <summary>
    /// Retrieve the first document matching a JSON containment query (<c>@&gt;</c>) ordered by the given fields in the
    /// document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns><c>Some</c> with the first document ordered by the given fields, or <c>None</c> if not found</returns>
    [<CompiledName "FSharpFirstByContainsOrdered">]
    let firstByContainsOrdered<'TDoc> tableName (criteria: obj) orderFields =
        WithProps.Find.firstByContainsOrdered<'TDoc> tableName criteria orderFields (fromDataSource ())

    /// <summary>
    /// Retrieve the first document matching a JSON containment query (<c>@&gt;</c>) ordered by the given fields in the
    /// document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>The first document ordered by the given fields, or <c>null</c> if not found</returns>
    let FirstByContainsOrdered<'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, criteria: obj, orderFields) =
        WithProps.Find.FirstByContainsOrdered<'TDoc>(tableName, criteria, orderFields, fromDataSource ())

    /// <summary>Retrieve the first document matching a JSON Path match query (<c>@?</c>)</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <returns><c>Some</c> with the first document, or <c>None</c> if not found</returns>
    [<CompiledName "FSharpFirstByJsonPath">]
    let firstByJsonPath<'TDoc> tableName jsonPath =
        WithProps.Find.firstByJsonPath<'TDoc> tableName jsonPath (fromDataSource ())

    /// <summary>Retrieve the first document matching a JSON Path match query (<c>@?</c>)</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <returns>The first document, or <c>null</c> if not found</returns>
    let FirstByJsonPath<'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, jsonPath) =
        WithProps.Find.FirstByJsonPath<'TDoc>(tableName, jsonPath, fromDataSource ())

    /// <summary>
    /// Retrieve the first document matching a JSON Path match query (<c>@?</c>) ordered by the given fields in the
    /// document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns><c>Some</c> with the first document ordered by the given fields, or <c>None</c> if not found</returns>
    [<CompiledName "FSharpFirstByJsonPathOrdered">]
    let firstByJsonPathOrdered<'TDoc> tableName jsonPath orderFields =
        WithProps.Find.firstByJsonPathOrdered<'TDoc> tableName jsonPath orderFields (fromDataSource ())

    /// <summary>
    /// Retrieve the first document matching a JSON Path match query (<c>@?</c>) ordered by the given fields in the
    /// document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>The first document ordered by the given fields, or <c>null</c> if not found</returns>
    let FirstByJsonPathOrdered<'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, jsonPath, orderFields) =
        WithProps.Find.FirstByJsonPathOrdered<'TDoc>(tableName, jsonPath, orderFields, fromDataSource ())


/// <summary>Commands to retrieve documents as JSON</summary>
[<RequireQualifiedAccess>]
module Json =

    /// <summary>Retrieve all documents in the given table as a JSON array</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <returns>All documents from the given table as a JSON array</returns>
    [<CompiledName "All">]
    let all tableName =
        WithProps.Json.all tableName (fromDataSource ())

    /// <summary>Write all documents in the given table to the given <c>StreamWriter</c></summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    [<CompiledName "WriteAll">]
    let writeAll tableName writer =
        WithProps.Json.writeAll tableName writer (fromDataSource ())

    /// <summary>
    /// Retrieve all documents in the given table as a JSON array, ordered by the given fields in the document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents from the given table as a JSON array, ordered by the given fields</returns>
    [<CompiledName "AllOrdered">]
    let allOrdered tableName orderFields =
        WithProps.Json.allOrdered tableName orderFields (fromDataSource ())

    /// <summary>
    /// Write all documents in the given table to the given <c>StreamWriter</c>, ordered by the given fields in the
    /// document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    [<CompiledName "WriteAllOrdered">]
    let writeAllOrdered tableName writer orderFields =
        WithProps.Json.writeAllOrdered tableName writer orderFields (fromDataSource ())

    /// <summary>Retrieve a JSON document by its ID</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="docId">The ID of the document to retrieve</param>
    /// <returns>The JSON document if found, an empty JSON document otherwise</returns>
    [<CompiledName "ById">]
    let byId<'TKey> tableName (docId: 'TKey) =
        WithProps.Json.byId tableName docId (fromDataSource ())

    /// <summary>Write a JSON document to the given <c>StreamWriter</c> by its ID</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="docId">The ID of the document to retrieve</param>
    [<CompiledName "WriteById">]
    let writeById<'TKey> tableName writer (docId: 'TKey) =
        WithProps.Json.writeById tableName writer docId (fromDataSource ())

    /// <summary>Retrieve JSON documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>All JSON documents matching the given fields</returns>
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields =
        WithProps.Json.byFields tableName howMatched fields (fromDataSource ())

    /// <summary>
    /// Write JSON documents to the given <c>StreamWriter</c> matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    [<CompiledName "WriteByFields">]
    let writeByFields tableName writer howMatched fields =
        WithProps.Json.writeByFields tableName writer howMatched fields (fromDataSource ())

    /// <summary>
    /// Retrieve JSON documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given fields
    /// in the document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All JSON documents matching the given fields, ordered by the other given fields</returns>
    [<CompiledName "ByFieldsOrdered">]
    let byFieldsOrdered tableName howMatched queryFields orderFields =
        WithProps.Json.byFieldsOrdered tableName howMatched queryFields orderFields (fromDataSource ())

    /// <summary>
    /// Write JSON documents to the given <c>StreamWriter</c> matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)
    /// ordered by the given fields in the document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    [<CompiledName "WriteByFieldsOrdered">]
    let writeByFieldsOrdered tableName writer howMatched queryFields orderFields =
        WithProps.Json.writeByFieldsOrdered tableName writer howMatched queryFields orderFields (fromDataSource ())

    /// <summary>Retrieve JSON documents matching a JSON containment query (<c>@&gt;</c>)</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <returns>All JSON documents matching the given containment query</returns>
    [<CompiledName "ByContains">]
    let byContains tableName (criteria: obj) =
        WithProps.Json.byContains tableName criteria (fromDataSource ())

    /// <summary>
    /// Write JSON documents to the given <c>StreamWriter</c> matching a JSON containment query (<c>@&gt;</c>)
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="criteria">The document to match with the containment query</param>
    [<CompiledName "WriteByContains">]
    let writeByContains tableName writer (criteria: obj) =
        WithProps.Json.writeByContains tableName writer criteria (fromDataSource ())

    /// <summary>
    /// Retrieve JSON documents matching a JSON containment query (<c>@&gt;</c>) ordered by the given fields in the
    /// document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents matching the given containment query, ordered by the given fields</returns>
    [<CompiledName "ByContainsOrdered">]
    let byContainsOrdered tableName (criteria: obj) orderFields =
        WithProps.Json.byContainsOrdered tableName criteria orderFields (fromDataSource ())

    /// <summary>
    /// Write JSON documents to the given <c>StreamWriter</c> matching a JSON containment query (<c>@&gt;</c>) ordered
    /// by the given fields in the document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    [<CompiledName "WriteByContainsOrdered">]
    let writeByContainsOrdered tableName writer (criteria: obj) orderFields =
        WithProps.Json.writeByContainsOrdered tableName writer criteria orderFields (fromDataSource ())

    /// <summary>Retrieve JSON documents matching a JSON Path match query (<c>@?</c>)</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <returns>All JSON documents matching the given JSON Path expression</returns>
    [<CompiledName "ByJsonPath">]
    let byJsonPath tableName jsonPath =
        WithProps.Json.byJsonPath tableName jsonPath (fromDataSource ())

    /// <summary>
    /// Write JSON documents to the given <c>StreamWriter</c> matching a JSON Path match query (<c>@?</c>)
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    [<CompiledName "WriteByJsonPath">]
    let writeByJsonPath tableName writer jsonPath =
        WithProps.Json.writeByJsonPath tableName writer jsonPath (fromDataSource ())

    /// <summary>
    /// Retrieve JSON documents matching a JSON Path match query (<c>@?</c>) ordered by the given fields in the document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All JSON documents matching the given JSON Path expression, ordered by the given fields</returns>
    [<CompiledName "ByJsonPathOrdered">]
    let byJsonPathOrdered tableName jsonPath orderFields =
        WithProps.Json.byJsonPathOrdered tableName jsonPath orderFields (fromDataSource ())

    /// <summary>
    /// Write JSON documents to the given <c>StreamWriter</c> matching a JSON Path match query (<c>@?</c>) ordered by
    /// the given fields in the document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    [<CompiledName "WriteByJsonPathOrdered">]
    let writeByJsonPathOrdered tableName writer jsonPath orderFields =
        WithProps.Json.writeByJsonPathOrdered tableName writer jsonPath orderFields (fromDataSource ())

    /// <summary>Retrieve the first JSON document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>The first matching JSON document if found, an empty JSON document otherwise</returns>
    [<CompiledName "FirstByFields">]
    let firstByFields tableName howMatched fields =
        WithProps.Json.firstByFields tableName howMatched fields (fromDataSource ())

    /// <summary>
    /// Write the first JSON document to the given <c>StreamWriter</c> matching JSON field comparisons
    /// (<c>-&gt;&gt; =</c>, etc.)
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    [<CompiledName "WriteFirstByFields">]
    let writeFirstByFields tableName writer howMatched fields =
        WithProps.Json.writeFirstByFields tableName writer howMatched fields (fromDataSource ())

    /// <summary>
    /// Retrieve the first JSON document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given
    /// fields in the document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>The first matching JSON document if found, an empty JSON document otherwise</returns>
    [<CompiledName "FirstByFieldsOrdered">]
    let firstByFieldsOrdered tableName howMatched queryFields orderFields =
        WithProps.Json.firstByFieldsOrdered tableName howMatched queryFields orderFields (fromDataSource ())

    /// <summary>
    /// Write the first JSON document to the given <c>StreamWriter</c> matching JSON field comparisons
    /// (<c>-&gt;&gt; =</c>, etc.) ordered by the given fields in the document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    [<CompiledName "WriteFirstByFieldsOrdered">]
    let writeFirstByFieldsOrdered tableName writer howMatched queryFields orderFields =
        WithProps.Json.writeFirstByFieldsOrdered tableName writer howMatched queryFields orderFields (fromDataSource ())

    /// <summary>Retrieve the first JSON document matching a JSON containment query (<c>@&gt;</c>)</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <returns>The first matching JSON document if found, an empty JSON document otherwise</returns>
    [<CompiledName "FirstByContains">]
    let firstByContains tableName (criteria: obj) =
        WithProps.Json.firstByContains tableName criteria (fromDataSource ())

    /// <summary>
    /// Write the first JSON document to the given <c>StreamWriter</c> matching a JSON containment query (<c>@&gt;</c>)
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="criteria">The document to match with the containment query</param>
    [<CompiledName "WriteFirstByContains">]
    let writeFirstByContains tableName writer (criteria: obj) =
        WithProps.Json.writeFirstByContains tableName writer criteria (fromDataSource ())

    /// <summary>
    /// Retrieve the first JSON document matching a JSON containment query (<c>@&gt;</c>) ordered by the given fields in
    /// the document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>The first matching JSON document if found, an empty JSON document otherwise</returns>
    [<CompiledName "FirstByContainsOrdered">]
    let firstByContainsOrdered tableName (criteria: obj) orderFields =
        WithProps.Json.firstByContainsOrdered tableName criteria orderFields (fromDataSource ())

    /// <summary>
    /// Write the first JSON document to the given <c>StreamWriter</c> matching a JSON containment query (<c>@&gt;</c>)
    /// ordered by the given fields in the document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="criteria">The document to match with the containment query</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    [<CompiledName "WriteFirstByContainsOrdered">]
    let writeFirstByContainsOrdered tableName writer (criteria: obj) orderFields =
        WithProps.Json.writeFirstByContainsOrdered tableName writer criteria orderFields (fromDataSource ())

    /// <summary>Retrieve the first JSON document matching a JSON Path match query (<c>@?</c>)</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <returns>The first matching JSON document if found, an empty JSON document otherwise</returns>
    [<CompiledName "FirstByJsonPath">]
    let firstByJsonPath tableName jsonPath =
        WithProps.Json.firstByJsonPath tableName jsonPath (fromDataSource ())

    /// <summary>
    /// Write the first JSON document to the given <c>StreamWriter</c> matching a JSON Path match query (<c>@?</c>)
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    [<CompiledName "WriteFirstByJsonPath">]
    let writeFirstByJsonPath tableName writer jsonPath =
        WithProps.Json.writeFirstByJsonPath tableName writer jsonPath (fromDataSource ())

    /// <summary>
    /// Retrieve the first JSON document matching a JSON Path match query (<c>@?</c>) ordered by the given fields in the
    /// document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>The first matching JSON document if found, an empty JSON document otherwise</returns>
    [<CompiledName "FirstByJsonPathOrdered">]
    let firstByJsonPathOrdered tableName jsonPath orderFields =
        WithProps.Json.firstByJsonPathOrdered tableName jsonPath orderFields (fromDataSource ())

    /// <summary>
    /// Write the first JSON document to the given <c>StreamWriter</c> matching a JSON Path match query (<c>@?</c>)
    /// ordered by the given fields in the document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    [<CompiledName "WriteFirstByJsonPathOrdered">]
    let writeFirstByJsonPathOrdered tableName writer jsonPath orderFields =
        WithProps.Json.writeFirstByJsonPathOrdered tableName writer jsonPath orderFields (fromDataSource ())


/// <summary>Commands to update documents</summary>
[<RequireQualifiedAccess>]
module Update =

    /// <summary>Update (replace) an entire document by its ID</summary>
    /// <param name="tableName">The table in which a document should be updated (may include schema)</param>
    /// <param name="docId">The ID of the document to be updated (replaced)</param>
    /// <param name="document">The new document</param>
    [<CompiledName "ById">]
    let byId tableName (docId: 'TKey) (document: 'TDoc) =
        WithProps.Update.byId tableName docId document (fromDataSource ())

    /// <summary>
    /// Update (replace) an entire document by its ID, using the provided function to obtain the ID from the document
    /// </summary>
    /// <param name="tableName">The table in which a document should be updated (may include schema)</param>
    /// <param name="idFunc">The function to obtain the ID of the document</param>
    /// <param name="document">The new document</param>
    [<CompiledName "FSharpFullFunc">]
    let byFunc tableName (idFunc: 'TDoc -> 'TKey) (document: 'TDoc) =
        WithProps.Update.byFunc tableName idFunc document (fromDataSource ())

    /// <summary>
    /// Update (replace) an entire document by its ID, using the provided function to obtain the ID from the document
    /// </summary>
    /// <param name="tableName">The table in which a document should be updated (may include schema)</param>
    /// <param name="idFunc">The function to obtain the ID of the document</param>
    /// <param name="document">The new document</param>
    let ByFunc(tableName, idFunc: System.Func<'TDoc, 'TKey>, document: 'TDoc) =
        WithProps.Update.ByFunc(tableName, idFunc, document, fromDataSource ())


/// <summary>Commands to patch (partially update) documents</summary>
[<RequireQualifiedAccess>]
module Patch =

    /// <summary>Patch a document by its ID</summary>
    /// <param name="tableName">The table in which a document should be patched (may include schema)</param>
    /// <param name="docId">The ID of the document to patch</param>
    /// <param name="patch">The partial document to patch the existing document</param>
    [<CompiledName "ById">]
    let byId tableName (docId: 'TKey) (patch: 'TPatch) =
        WithProps.Patch.byId tableName docId patch (fromDataSource ())

    /// <summary>
    /// Patch documents using a JSON field comparison query in the <c>WHERE</c> clause (<c>-&gt;&gt; =</c>, etc.)
    /// </summary>
    /// <param name="tableName">The table in which documents should be patched (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <param name="patch">The partial document to patch the existing document</param>
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields (patch: 'TPatch) =
        WithProps.Patch.byFields tableName howMatched fields patch (fromDataSource ())

    /// <summary>Patch documents using a JSON containment query in the <c>WHERE</c> clause (<c>@&gt;</c>)</summary>
    /// <param name="tableName">The table in which documents should be patched (may include schema)</param>
    /// <param name="criteria">The document to match the containment query</param>
    /// <param name="patch">The partial document to patch the existing document</param>
    [<CompiledName "ByContains">]
    let byContains tableName (criteria: 'TCriteria) (patch: 'TPatch) =
        WithProps.Patch.byContains tableName criteria patch (fromDataSource ())

    /// <summary>Patch documents using a JSON Path match query in the <c>WHERE</c> clause (<c>@?</c>)</summary>
    /// <param name="tableName">The table in which documents should be patched (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <param name="patch">The partial document to patch the existing document</param>
    [<CompiledName "ByJsonPath">]
    let byJsonPath tableName jsonPath (patch: 'TPatch) =
        WithProps.Patch.byJsonPath tableName jsonPath patch (fromDataSource ())


/// <summary>Commands to remove fields from documents</summary>
[<RequireQualifiedAccess>]
module RemoveFields =

    /// <summary>Remove fields from a document by the document's ID</summary>
    /// <param name="tableName">The table in which a document should be modified (may include schema)</param>
    /// <param name="docId">The ID of the document to modify</param>
    /// <param name="fieldNames">One or more field names to remove from the document</param>
    [<CompiledName "ById">]
    let byId tableName (docId: 'TKey) fieldNames =
        WithProps.RemoveFields.byId tableName docId fieldNames (fromDataSource ())

    /// <summary>Remove fields from documents via a comparison on JSON fields in the document</summary>
    /// <param name="tableName">The table in which documents should be modified (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <param name="fieldNames">One or more field names to remove from the matching documents</param>
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields fieldNames =
        WithProps.RemoveFields.byFields tableName howMatched fields fieldNames (fromDataSource ())

    /// <summary>Remove fields from documents via a JSON containment query (<c>@&gt;</c>)</summary>
    /// <param name="tableName">The table in which documents should be modified (may include schema)</param>
    /// <param name="criteria">The document to match the containment query</param>
    /// <param name="fieldNames">One or more field names to remove from the matching documents</param>
    [<CompiledName "ByContains">]
    let byContains tableName (criteria: 'TContains) fieldNames =
        WithProps.RemoveFields.byContains tableName criteria fieldNames (fromDataSource ())

    /// <summary>Remove fields from documents via a JSON Path match query (<c>@?</c>)</summary>
    /// <param name="tableName">The table in which documents should be modified (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    /// <param name="fieldNames">One or more field names to remove from the matching documents</param>
    [<CompiledName "ByJsonPath">]
    let byJsonPath tableName jsonPath fieldNames =
        WithProps.RemoveFields.byJsonPath tableName jsonPath fieldNames (fromDataSource ())


/// <summary>Commands to delete documents</summary>
[<RequireQualifiedAccess>]
module Delete =

    /// <summary>Delete a document by its ID</summary>
    /// <param name="tableName">The table in which a document should be deleted (may include schema)</param>
    /// <param name="docId">The ID of the document to delete</param>
    [<CompiledName "ById">]
    let byId tableName (docId: 'TKey) =
        WithProps.Delete.byId tableName docId (fromDataSource ())

    /// <summary>Delete documents by matching a JSON field comparison query (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table in which documents should be deleted (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields =
        WithProps.Delete.byFields tableName howMatched fields (fromDataSource ())

    /// <summary>Delete documents by matching a JSON contains query (<c>@&gt;</c>)</summary>
    /// <param name="tableName">The table in which documents should be deleted (may include schema)</param>
    /// <param name="criteria">The document to match the containment query</param>
    [<CompiledName "ByContains">]
    let byContains tableName (criteria: 'TContains) =
        WithProps.Delete.byContains tableName criteria (fromDataSource ())

    /// <summary>Delete documents by matching a JSON Path match query (@?)</summary>
    /// <param name="tableName">The table in which documents should be deleted (may include schema)</param>
    /// <param name="jsonPath">The JSON Path expression to match</param>
    [<CompiledName "ByJsonPath">]
    let byJsonPath tableName jsonPath =
        WithProps.Delete.byJsonPath tableName jsonPath (fromDataSource ())
